import json
import os
import shutil
import tempfile

import mock
import pytest
from click.testing import CliRunner

from gradient.api_sdk import sdk_exceptions
from gradient.api_sdk.clients.http_client import default_headers
from gradient.cli import cli
from tests import example_responses, MockResponse
from tests.example_responses import LIST_JOB_FILES_RESPONSE_JSON

EXPECTED_HEADERS = default_headers.copy()
EXPECTED_HEADERS["ps_client_name"] = "gradient-cli"

EXPECTED_HEADERS_WITH_CHANGED_API_KEY = EXPECTED_HEADERS.copy()
EXPECTED_HEADERS_WITH_CHANGED_API_KEY["X-API-Key"] = "some_key"


@pytest.fixture
def basic_options_metrics_stream_websocket_connection_iterator():
    def generator(self):
        yield """{"handle": "jstkd2lapucirs", "object_type": "mljob", "chart_name": "memoryUsage",
               "pod_metrics": {"mljob-ecgrgm7ok8chv-0-worker": {"time_stamp": 1588155670, "value": "5881856"}}}"""

        yield """{"handle": "jstkd2lapucirs", "object_type": "mljob", "chart_name": "cpuPercentage",
               "pod_metrics": {"mljob-ecgrgm7ok8chv-0-worker": {"time_stamp": 1588155670, "value": "0"}}}"""

        yield """{"handle": "jstkd2lapucirs", "object_type": "mljob", "chart_name": "memoryUsage",
               "pod_metrics": {"mljob-ecgrgm7ok8chv-0-worker": {"time_stamp": 1588155700, "value": "5881857"}}}"""

        raise sdk_exceptions.EndWebsocketStream("keton")

    return generator


@pytest.fixture
def all_options_metrics_stream_websocket_connection_iterator():
    def generator(self):
        yield """{"handle": "jstkd2lapucirs", "object_type": "mljob", "chart_name": "gpuMemoryUsed",
               "pod_metrics": {"mljob-ecgrgm7ok8chv-0-worker": {"time_stamp": 1588155670, "value": "5881856"}}}"""

        yield """{"handle": "jstkd2lapucirs", "object_type": "mljob", "chart_name": "gpuMemoryFree",
               "pod_metrics": {"mljob-ecgrgm7ok8chv-0-worker": {"time_stamp": 1588155670, "value": "0"}}}"""

        yield """{"handle": "jstkd2lapucirs", "object_type": "mljob", "chart_name": "gpuMemoryUsed",
               "pod_metrics": {"mljob-ecgrgm7ok8chv-0-worker": {"time_stamp": 1588155700, "value": "5881857"}}}"""

        raise sdk_exceptions.EndWebsocketStream("keton")

    return generator


class TestJobs(object):
    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to fetch data: Invalid API token\n"
    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}

    EXPECTED_HEADERS = default_headers.copy()
    EXPECTED_HEADERS_WITH_CHANGED_API_KEY = default_headers.copy()
    EXPECTED_HEADERS_WITH_CHANGED_API_KEY["X-API-Key"] = "some_key"


class TestListJobs(TestJobs):
    URL = "https://api.paperspace.io/jobs/getJobList/"
    BASIC_COMMAND = ["jobs", "list"]
    COMMAND_WITH_OPTIONS_FILE = ["jobs", "list", "--optionsFile", ]  # path added in test

    EXPECTED_RESPONSE_JSON = example_responses.LIST_JOBS_RESPONSE_JSON
    EXPECTED_STDOUT = """+----------------+---------------------------+-------------------+----------------+--------------+--------------------------+
| ID             | Name                      | Project           | Cluster        | Machine Type | Created                  |
+----------------+---------------------------+-------------------+----------------+--------------+--------------------------+
| jsxeeba5qq99yn | job 1                     | keton             | PS Jobs on GCP | K80          | 2019-03-25T14:51:16.118Z |
| jfl063dsv634h  | job 2                     | keton             | PS Jobs on GCP | P100         | 2019-03-25T14:54:30.866Z |
| jsvau8w47k78zm | Clone - jfl063dsv634h     | keton             | PS Jobs on GCP | P100         | 2019-03-25T15:04:43.844Z |
| j2eq99xhvgtum  | keton1-worker-1           | keton             | PS Jobs on GCP | P100         | 2019-03-25T15:07:30.383Z |
| jzzinybinuxf9  | keton2-worker-1           | keton             | PS Jobs on GCP | P100         | 2019-03-25T15:18:51.461Z |
| jsb37duc1zlbz0 | keton4-worker-1           | keton             | PS Jobs on GCP | P100         | 2019-03-25T15:29:04.601Z |
| jq41vipwy18f7  | keton4-parameter_server-1 | keton             | PS Jobs on GCP | P100         | 2019-03-25T15:29:06.765Z |
| jsigkjnyb6m3qm | Test1-worker-1            | keton             | PS Jobs on GCP | K80          | 2019-04-02T15:17:05.618Z |
| j4g76vuppxqao  | job 1                     | paperspace-python | PS Jobs on GCP | K80          | 2019-04-04T15:12:34.414Z |
| jsbnvdhwb46vr9 | job 2                     | paperspace-python | PS Jobs on GCP | G1           | 2019-04-24T09:09:53.645Z |
| jt8alwzv28kha  | job 3                     | paperspace-python | PS Jobs on GCP | G1           | 2019-04-24T10:18:30.620Z |
+----------------+---------------------------+-------------------+----------------+--------------+--------------------------+
"""

    BASIC_COMMAND_WITH_API_KEY = ["jobs", "list", "--apiKey", "some_key"]

    RESPONSE_JSON_WHEN_NO_JOBS_WERE_FOUND = []
    EXPECTED_STDOUT_WHEN_NO_JOBS_WERE_FOUND = "No data found\n"

    BASIC_COMMAND_WITH_FILTERING = [
        "jobs", "list",
        "--projectId", "some_project_id",
        "--tag", "some_tag",
        "--tag", "some_other_tag",
    ]
    EXPECTED_PARAMS_WITHOUT_FILTERING = {
        "filter": "{\"filter\": {\"where\": {}}}",
    }
    EXPECTED_REQUEST_PARAMS_WITH_FILTERING = {
        "filter": "{\"filter\": {\"where\": {\"projectId\": \"some_project_id\"}}}",
        "modelName": "team",
        "tagFilter[0]": "some_tag",
        "tagFilter[1]": "some_other_tag",
    }

    BASIC_COMMAND_WITH_MUTUALLY_EXCLUSIVE_FILTERS = [
        "jobs", "list",
        "--project", "some_project_name",
        "--projectId", "some_project_id",
    ]

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_and_print_table_when_jobs_list_was_used(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.EXPECTED_PARAMS_WITHOUT_FILTERING)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_jobs_list_was_used_with_api_key_option(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.EXPECTED_PARAMS_WITHOUT_FILTERING)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(self, get_patched, jobs_list_config_path):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [jobs_list_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.EXPECTED_REQUEST_PARAMS_WITH_FILTERING)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_jobs_list_was_used_with_wrong_api_key(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WITH_WRONG_API_TOKEN, status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_API_KEY)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.EXPECTED_PARAMS_WITHOUT_FILTERING)
        assert result.output == self.EXPECTED_STDOUT_WITH_WRONG_API_TOKEN
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_no_job_was_not_found(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.RESPONSE_JSON_WHEN_NO_JOBS_WERE_FOUND)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.EXPECTED_PARAMS_WITHOUT_FILTERING)
        assert result.output == self.EXPECTED_STDOUT_WHEN_NO_JOBS_WERE_FOUND
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_error_status_code_received_but_no_content_was_provided(self, get_patched):
        get_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.EXPECTED_PARAMS_WITHOUT_FILTERING)
        assert result.output == "Failed to fetch data\n"
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_post_request_when_jobs_list_was_used_with_filter_options(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON, status_code=200)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.BASIC_COMMAND_WITH_FILTERING)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS,
                                       json=None,
                                       params=self.EXPECTED_REQUEST_PARAMS_WITH_FILTERING)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0


class TestJobLogs(TestJobs):
    URL = "https://logs.paperspace.io/jobs/logs"

    RESPONSE_JSON_WITH_WRONG_API_TOKEN = {"status": 400, "message": "Invalid API token"}
    EXPECTED_RESPONSE_JSON = example_responses.LIST_OF_LOGS_FOR_JOB
    COMMAND_WITHOUT_REQUIRED_PARAMETERS = ["jobs", "logs"]
    COMMAND_WITH_ALL_OPTIONS = [
        "jobs", "logs",
        "--id", "some_id",
        "--apiKey", "some_key",
        "--line", "50",
        "--limit", "200",
        # "--follow", "True",
    ]

    COMMAND_WITH_OPTIONS_FILE = ["jobs", "logs", "--optionsFile", ]  # path added in test
    REQUEST_WITH_OPTIONS_FILE = {"jobId": "some_id", "line": 50, "limit": 200}

    EXPECTED_STDOUT = """+Job some_id logs------------------------------------------------------------------------+
| LINE | MESSAGE                                                                         |
+------+---------------------------------------------------------------------------------+
| 1    | Traceback (most recent call last):                                              |
| 2    |   File "generate_figures.py", line 15, in <module>                              |
| 3    |     import dnnlib.tflib as tflib                                                |
| 4    |   File "/paperspace/dnnlib/tflib/__init__.py", line 8, in <module>              |
| 5    |     from . import autosummary                                                   |
| 6    |   File "/paperspace/dnnlib/tflib/autosummary.py", line 31, in <module>          |
| 7    |     from . import tfutil                                                        |
| 8    |   File "/paperspace/dnnlib/tflib/tfutil.py", line 34, in <module>               |
| 9    |     def shape_to_list(shape: Iterable[tf.Dimension]) -> List[Union[int, None]]: |
| 10   | AttributeError: module \'tensorflow\' has no attribute 'Dimension'                |
| 11   | PSEOF                                                                           |
+------+---------------------------------------------------------------------------------+
"""

    EXPECTED_STDOUT_WITH_WRONG_API_TOKEN = "Failed to fetch data: Invalid API token\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_get_request_and_print_available_logs(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND_WITH_ALL_OPTIONS)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.REQUEST_WITH_OPTIONS_FILE)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_get_request_when_log_list_was_used_with_wrong_api_key(self, get_patched):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND_WITH_ALL_OPTIONS)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.REQUEST_WITH_OPTIONS_FILE)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(self, get_patched, jobs_logs_config_path):
        get_patched.return_value = MockResponse(json_data=self.EXPECTED_RESPONSE_JSON)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [jobs_logs_config_path]

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, command)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.REQUEST_WITH_OPTIONS_FILE)
        assert result.output == self.EXPECTED_STDOUT
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_error_message_when_error_status_code_received_but_no_content_was_provided(self, get_patched):
        get_patched.return_value = MockResponse(status_code=400)

        cli_runner = CliRunner()
        result = cli_runner.invoke(cli.cli, self.COMMAND_WITH_ALL_OPTIONS)

        get_patched.assert_called_with(self.URL,
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params=self.REQUEST_WITH_OPTIONS_FILE)
        assert result.output == "Failed to fetch data\n"
        assert result.exit_code == 0


class TestDestroyJobArtifactsCommands(TestJobs):
    runner = CliRunner()
    URL = "https://api.paperspace.io"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_post_request_when_destroying_artifacts_with_files_specified(self, post_patched):
        post_patched.return_value = MockResponse(status_code=200)
        job_id = "some_job_id"
        file_names = "some_file_names"
        result = self.runner.invoke(cli.cli, ["jobs", "artifacts", "destroy", "--id", job_id, "--files", file_names,
                                              "--apiKey", "some_key"])

        assert result.exit_code == 0, result.exc_info
        post_patched.assert_called_with("{}/jobs/{}/artifactsDestroy".format(self.URL, job_id),
                                        files=None,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=None,
                                        params={"files": file_names},
                                        data=None)

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_valid_post_request_when_destroying_artifacts_without_files_specified(self, post_patched):
        post_patched.return_value = MockResponse(status_code=200)
        job_id = "some_job_id"
        result = self.runner.invoke(cli.cli, ["jobs", "artifacts", "destroy", "--id", job_id, "--apiKey", "some_key"])

        post_patched.assert_called_with("{}/jobs/{}/artifactsDestroy".format(self.URL, job_id),
                                        files=None,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=None,
                                        params=None,
                                        data=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_options_from_yaml_file(self, post_patched, jobs_artifacts_destroy_config_path):
        post_patched.return_value = MockResponse()
        command = ["jobs", "artifacts", "destroy", "--optionsFile", jobs_artifacts_destroy_config_path]

        result = self.runner.invoke(cli.cli, command)

        assert result.exit_code == 0, result.exc_info
        post_patched.assert_called_with("{}/jobs/some_id/artifactsDestroy".format(self.URL),
                                        files=None,
                                        headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                        json=None,
                                        params={"files": "file1,file2"},
                                        data=None)


class TestGetJobArtifacts(TestJobs):
    runner = CliRunner()
    URL = "https://api.paperspace.io"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_send_valid_get_request_and_receive_json_response(self, get_patched):
        get_patched.return_value = MockResponse()
        job_id = "some_job_id"
        result = self.runner.invoke(cli.cli, ["jobs", "artifacts", "get", "--id", job_id, "--apiKey", "some_key"])

        get_patched.assert_called_with("{}/jobs/artifactsGet".format(self.URL),
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params={"jobId": job_id})
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(self, get_patched, jobs_artifacts_get_config_path):
        get_patched.return_value = MockResponse()
        command = ["jobs", "artifacts", "get", "--optionsFile", jobs_artifacts_get_config_path]

        result = self.runner.invoke(cli.cli, command)

        get_patched.assert_called_with("{}/jobs/artifactsGet".format(self.URL),
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params={"jobId": "some_id"})
        assert result.exit_code == 0


class TestListJobArtifacts(TestJobs):
    runner = CliRunner()
    URL = "https://api.paperspace.io"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_valid_get_request_with_all_parameters_for_a_list_of_artifacts(self, get_patched):
        get_patched.return_value = MockResponse(LIST_JOB_FILES_RESPONSE_JSON)
        job_id = "some_job_id"
        result = self.runner.invoke(cli.cli,
                                    ["jobs", "artifacts", "list", "--id", job_id, "--apiKey", "some_key", "--size",
                                     "--links",
                                     "--files", "foo"])

        get_patched.assert_called_with("{}/jobs/artifactsListV2".format(self.URL),
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params={"jobId": job_id,
                                               "size": True,
                                               "links": True,
                                               "files": "foo"})
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(self, get_patched, jobs_artifacts_list_config_path):
        get_patched.return_value = MockResponse(LIST_JOB_FILES_RESPONSE_JSON)
        command = ["jobs", "artifacts", "list", "--optionsFile", jobs_artifacts_list_config_path]
        result = self.runner.invoke(cli.cli, command)

        get_patched.assert_called_with("{}/jobs/artifactsListV2".format(self.URL),
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params={"files": "keton*.py",
                                               "size": True,
                                               "links": True,
                                               "jobId": "some_id"})
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    @pytest.mark.parametrize('option,param', [("--size", "size"),
                                              ("-s", "size"),
                                              ("--links", "links"),
                                              ("-l", "links")])
    def test_should_send_valid_get_request_with_valid_param_for_a_list_of_artifacts_for_both_formats_of_param(self,
                                                                                                              get_patched,
                                                                                                              option,
                                                                                                              param):
        get_patched.return_value = MockResponse(LIST_JOB_FILES_RESPONSE_JSON, status_code=200)
        job_id = "some_job_id"
        result = self.runner.invoke(cli.cli,
                                    ["jobs", "artifacts", "list", "--id", job_id, "--apiKey", "some_key"] + [option])

        get_patched.assert_called_with("{}/jobs/artifactsListV2".format(self.URL),
                                       headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                       json=None,
                                       params={"jobId": job_id,
                                               param: True})
        assert result.exit_code == 0


class TestJobsCreate(object):
    URL = "https://api.paperspace.io"
    TAGS_URL = "https://api.paperspace.io/entityTags/updateTags"
    EXPECTED_HEADERS = default_headers.copy()
    EXPECTED_HEADERS_WITH_CHANGED_API_KEY = default_headers.copy()
    EXPECTED_HEADERS_WITH_CHANGED_API_KEY["X-API-Key"] = "some_key"
    BASIC_OPTIONS_COMMAND = [
        "jobs", "create",
        "--name", "exp1",
        "--projectId", "testHandle",
        "--container", "testContainer",
        "--machineType", "testType",
        "--command", "testCommand",
        "--workspace", "https://github.com/Paperspace/gradient-cli.git",
    ]
    BASIC_OPTIONS_COMMAND_WITH_TAGS = [
        "jobs", "create",
        "--name", "exp1",
        "--projectId", "testHandle",
        "--container", "testContainer",
        "--machineType", "testType",
        "--command", "testCommand",
        "--workspace", "https://github.com/Paperspace/gradient-cli.git",
        "--tag", "test0",
        "--tag", "test1",
        "--tags", "test2,test3",
    ]
    FULL_OPTIONS_COMMAND = [
        "jobs", "create",
        "--apiKey", "some_key",
        "--cluster", "some_cluster_name",
        "--clusterId", "some_cluster_id",
        "--command", "some command",
        "--container", "some_container",
        "--experimentId", "some_experiment_id",
        "--ignoreFiles", "file1,file2",
        "--isPreemptible", "True",
        "--projectId", "some_project_id",
        "--isPublic", "True",
        "--envVars", '{"key":"val"}',
        "--machineType", "K80",
        "--name", "some_name",
        "--nodeAttrs", '{"key":"val"}',
        "--ports", "8080,9000:9900",
        "--projectId", "some_project_id",
        "--registryPassword", "some_registry_password",
        "--registryUsername", "some_registry_username",
        "--registryTarget", "some_registry_target",
        "--registryTargetPassword", "some_registry_target_password",
        "--registryTargetUsername", "some_registry_target_username",
        "--relDockerfilePath", "some dockerfile path",
        "--startedByUserId", "some_user_id",
        "--useDockerfile", "True",
        "--workingDirectory", "/some/path",
        "--buildOnly",
        "--workspace", "s3://some-path",
    ]
    BASIC_OPTIONS_REQUEST = {
        "name": u"exp1",
        "projectId": u"testHandle",
        "container": u"testContainer",
        "machineType": u"testType",
        "command": u"testCommand",
        "workspaceFileName": u"https://github.com/Paperspace/gradient-cli.git",
    }
    FULL_OPTIONS_REQUEST = {
        "clusterId": "some_cluster_id",
        "experimentId": "some_experiment_id",
        "cluster": "some_cluster_name",
        "startedByUserId": "some_user_id",
        "isPreemptible": True,
        "container": "some_container",
        "workingDirectory": "/some/path",
        "projectId": "some_project_id",
        "registryTargetUsername": "some_registry_target_username",
        "machineType": "K80",
        "registryTargetPassword": "some_registry_target_password",
        "registryTarget": "some_registry_target",
        "isPublic": True,
        "workspaceFileName": "s3://some-path",
        "envVars": '{"key": "val"}',
        "useDockerfile": True,
        "name": "some_name",
        "relDockerfilePath": "some dockerfile path",
        "targetNodeAttrs": {"key": "val"},
        "command": "some command",
        "ports": "8080,9000:9900",
        "buildOnly": True,
        "registryUsername": "some_registry_username",
        "registryPassword": "some_registry_password",
    }
    TAGS_JSON = {
        "entity": "job",
        "entityId": "sadkfhlskdjh",
        "tags": ["test0", "test1", "test2", "test3"]
    }
    RESPONSE_JSON_200 = {"id": "sadkfhlskdjh", "message": "success"}
    UPDATE_TAGS_RESPONSE_JSON_200 = example_responses.UPDATE_TAGS_RESPONSE
    RESPONSE_CONTENT_200 = b'{"handle":"sadkfhlskdjh","message":"success"}\n'
    EXPECTED_STDOUT = u'New job created with ID: sadkfhlskdjh\n'
    EXPECTED_STDOUT_TAGS = u'New job created with ID: sadkfhlskdjh\n' \
                           u'https://console.paperspace.com/jobs/sadkfhlskdjh\n'

    RESPONSE_JSON_404_PROJECT_NOT_FOUND = {"details": {"handle": "wrong_handle"}, "error": "Project not found"}
    RESPONSE_CONTENT_404_PROJECT_NOT_FOUND = b'{"details":{"handle":"wrong_handle"},"error":"Project not found"}\n'
    EXPECTED_STDOUT_PROJECT_NOT_FOUND = "Project not found\nhandle: wrong_handle\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_proper_data_and_print_message_when_create_job_was_run_with_basic_options(self, post_patched):
        post_patched.return_value = MockResponse(self.RESPONSE_JSON_200, 200, self.RESPONSE_CONTENT_200)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.BASIC_OPTIONS_COMMAND)

        assert self.EXPECTED_STDOUT in result.output, result.exc_info
        post_patched.assert_called_once_with(self.URL + '/jobs/createJob/',
                                             headers=EXPECTED_HEADERS,
                                             json=None,
                                             params=self.BASIC_OPTIONS_REQUEST,
                                             files=None,
                                             data=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_proper_data_and_print_message_when_create_job_was_run_with_all_options(self, post_patched):
        post_patched.return_value = MockResponse(self.RESPONSE_JSON_200, 200, self.RESPONSE_CONTENT_200)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.FULL_OPTIONS_COMMAND)

        assert self.EXPECTED_STDOUT in result.output, result.exc_info
        post_patched.assert_called_once_with(self.URL + '/jobs/createJob/',
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=None,
                                             params=self.FULL_OPTIONS_REQUEST,
                                             files=None,
                                             data=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_options_from_yaml_file(self, post_patched, jobs_create_config_path):
        post_patched.return_value = MockResponse(self.RESPONSE_JSON_200, 200, self.RESPONSE_CONTENT_200)
        command = ["jobs", "create", "--optionsFile", jobs_create_config_path]

        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        assert self.EXPECTED_STDOUT in result.output, result.exc_info
        post_patched.assert_called_once_with(self.URL + '/jobs/createJob/',
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=None,
                                             params=self.FULL_OPTIONS_REQUEST,
                                             files=None,
                                             data=None)
        assert result.exit_code == 0

    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_proper_data_and_tag_job(self, post_patched, get_patched, put_patched):
        post_patched.return_value = MockResponse(self.RESPONSE_JSON_200, 200, self.RESPONSE_CONTENT_200)
        get_patched.return_value = MockResponse({}, 200, "fake content")
        put_patched.return_value = MockResponse(self.UPDATE_TAGS_RESPONSE_JSON_200, 200, "fake content")

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.BASIC_OPTIONS_COMMAND_WITH_TAGS)

        post_patched.assert_called_once_with(self.URL + '/jobs/createJob/',
                                             headers=EXPECTED_HEADERS,
                                             json=None,
                                             params=self.BASIC_OPTIONS_REQUEST,
                                             files=None,
                                             data=None)

        put_patched.assert_called_once_with(
            self.TAGS_URL,
            headers=EXPECTED_HEADERS,
            json=self.TAGS_JSON,
            params=None,
            data=None,
        )

        assert result.output == self.EXPECTED_STDOUT_TAGS, result.exc_info
        assert result.exit_code == 0


class TestDownloadJobArtifacts(TestJobs):
    runner = CliRunner()
    LIST_FILES_URL = "https://api.paperspace.io/jobs/artifactsListV2"
    DESTINATION_DIR_NAME = "dest"
    DESTINATION_DIR_PATH = os.path.join(tempfile.gettempdir(), "dest")

    COMMAND = ["jobs", "artifacts", "download", "--id", "some_job_id", "--destinationDir", DESTINATION_DIR_PATH]

    @classmethod
    def teardown_method(cls):
        shutil.rmtree(cls.DESTINATION_DIR_PATH)

    @mock.patch("gradient.api_sdk.s3_downloader.requests.get")
    def test_should_get_a_list_of_files_and_download_them_to_defined_directory_when_download_command_was_executed(
            self, get_patched,
    ):
        file_response_mock = mock.MagicMock()
        file_response_mock.content = "\"Hello Paperspace!\n\""
        file_response_mock_2 = mock.MagicMock()
        file_response_mock_2.content = "\"Hello Paperspace 2\n\""
        file_response_mock_3 = mock.MagicMock()
        file_response_mock_3.content = "\"Elo\n\""
        get_patched.side_effect = [
            MockResponse(LIST_JOB_FILES_RESPONSE_JSON),
            file_response_mock,
            file_response_mock_2,
            file_response_mock_3,
        ]

        result = self.runner.invoke(cli.cli, self.COMMAND)

        get_patched.assert_has_calls([
            mock.call(self.LIST_FILES_URL,
                      headers=EXPECTED_HEADERS,
                      json=None,
                      params={"links": True, "jobId": "some_job_id"}),
            mock.call("https://ps-projects.s3.amazonaws.com/some/path/artifacts/hello.txt?AWSAccessKeyId="
                      "some_aws_access_key_id&Expires=713274132&Signature=7CT5k6buEmZe5k5E7g6BXMs2xV4%3D&"
                      "response-content-disposition=attachment%3Bfilename%3D%22hello.txt%22&x-amz-security-token="
                      "some_amz_security_token"),
            mock.call("https://ps-projects.s3.amazonaws.com/some/path/artifacts/hello2.txt?AWSAccessKeyId="
                      "some_aws_access_key_id&Expires=713274132&Signature=L1lI47cNyiROzdYkf%2FF3Cm3165E%3D&"
                      "response-content-disposition=attachment%3Bfilename%3D%22hello2.txt%22&x-amz-security-token="
                      "some_amz_security_token"),
            mock.call("https://ps-projects.s3.amazonaws.com/some/path/artifacts/keton/elo.txt?AWSAccessKeyId="
                      "some_aws_access_key_id&Expires=713274132&Signature=tHriojGx03S%2FKkVGQGVI5CQRFTo%3D&"
                      "response-content-disposition=attachment%3Bfilename%3D%22elo.txt%22&x-amz-security-token="
                      "some_amz_security_token"),
        ])
        assert os.path.exists(self.DESTINATION_DIR_PATH)
        assert os.path.isdir(self.DESTINATION_DIR_PATH)
        assert os.path.exists(os.path.join(self.DESTINATION_DIR_PATH, "keton"))
        assert os.path.isdir(os.path.join(self.DESTINATION_DIR_PATH, "keton"))

        hello_txt_path = os.path.join(self.DESTINATION_DIR_PATH, "hello.txt")
        assert os.path.exists(hello_txt_path)
        assert not os.path.isdir(hello_txt_path)
        with open(hello_txt_path) as h:
            assert h.read() == "\"Hello Paperspace!\n\""

        hello2_txt_path = os.path.join(self.DESTINATION_DIR_PATH, "hello2.txt")
        assert os.path.exists(hello2_txt_path)
        assert not os.path.isdir(hello2_txt_path)
        with open(hello2_txt_path) as h:
            assert h.read() == "\"Hello Paperspace 2\n\""

        elo_txt_path = os.path.join(self.DESTINATION_DIR_PATH, "keton", "elo.txt")
        assert os.path.exists(elo_txt_path)
        assert not os.path.isdir(elo_txt_path)
        with open(elo_txt_path) as h:
            assert h.read() == "\"Elo\n\""

        assert result.exit_code == 0


class TestJobsMetricsGetCommand(object):
    GET_JOB_URL = "https://api.paperspace.io/jobs/getPublicJob"
    GET_METRICS_URL = "https://aws-testing.paperspace.io/metrics/api/v1/range"
    BASIC_OPTIONS_COMMAND = [
        "jobs", "metrics", "get",
        "--id", "jstkd2lapucirs",
    ]
    ALL_OPTIONS_COMMAND = [
        "jobs", "metrics", "get",
        "--id", "jstkd2lapucirs",
        "--metric", "gpuMemoryFree",
        "--metric", "gpuMemoryUsed",
        "--interval", "20s",
        "--start", "2020-04-01",
        "--end", "2020-04-02 21:37:00",
        "--apiKey", "some_key",
    ]
    FULL_OPTIONS_COMMAND_WITH_OPTIONS_FILE = [
        "jobs", "metrics", "get",
        "--optionsFile",  # path added in test,
    ]

    GET_JOB_REQUEST_JSON = {"jobId": "jstkd2lapucirs"}
    BASIC_COMMAND_GET_METRICS_REQUEST_PARAMS = {
        "start": "2020-04-29T10:11:07Z",
        "handle": "jstkd2lapucirs",
        "interval": "30s",
        "charts": "cpuPercentage,memoryUsage",
        "objecttype": "mljob",
    }
    ALL_COMMANDS_GET_METRICS_REQUEST_PARAMS = {
        "start": "2020-04-01T00:00:00Z",
        "handle": "jstkd2lapucirs",
        "interval": "20s",
        "charts": "gpuMemoryFree,gpuMemoryUsed",
        "objecttype": "mljob",
        "end": "2020-04-02T21:37:00Z",
    }

    GET_JOB_RESPONSE_JSON = example_responses.GET_JOB_RESPONSE
    GET_METRICS_RESPONSE_JSON = example_responses.GET_JOB_METRICS_RESPONSE
    GET_JOB_RESPONSE_JSON_WHEN_NO_JOBS_WERE_FOUND = {
        "error": {
            "name": "ApplicationError",
            "status": 404,
            "message": "No such job",
        },
    }

    EXPECTED_STDOUT = """{
  "cpuPercentage": {
    "mljob-ecgrgm7ok8chv-0-worker": [
      {
        "time_stamp": 1588155157, 
        "value": "0"
      }, 
      {
        "time_stamp": 1588155187, 
        "value": "0"
      }, 
      {
        "time_stamp": 1588155217, 
        "value": "0"
      }
    ]
  }, 
  "memoryUsage": {
    "mljob-ecgrgm7ok8chv-0-worker": [
      {
        "time_stamp": 1588155097, 
        "value": "0"
      }, 
      {
        "time_stamp": 1588155127, 
        "value": "5881856"
      }, 
      {
        "time_stamp": 1588155157, 
        "value": "5881856"
      }, 
      {
        "time_stamp": 1588155187, 
        "value": "5881856"
      }
    ]
  }
}
"""

    EXPECTED_STDOUT_WHEN_INVALID_API_KEY_WAS_USED = "Failed to fetch data: Invalid API token\n"
    EXPECTED_STDOUT_WHEN_EXPERIMENT_WAS_NOT_FOUND = "Failed to fetch data: No such job\n"
    EXPECTED_STDOUT_WHEN_NO_METRICS_FOUND = """{
    "cpuPercentage": null,
    "memoryUsage": null
}
"""
    EXPECTED_STDOUT_WHEN_ERROR_CODE_WAS_RETURNED_WITHOUT_ERROR_MESSAGE = "Failed to fetch data\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_all_available_metrics_when_metrics_get_command_was_used_with_basic_options(
            self, post_patched, get_patched):
        post_patched.return_value = MockResponse(self.GET_JOB_RESPONSE_JSON)
        get_patched.return_value = MockResponse(self.GET_METRICS_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.BASIC_OPTIONS_COMMAND)

        assert json.loads(result.output.strip()) == json.loads(self.EXPECTED_STDOUT.strip()), \
            str(result.output) + str(result.exc_info)
        post_patched.assert_called_once_with(
            self.GET_JOB_URL,
            json=self.GET_JOB_REQUEST_JSON,
            params=None,
            data=None,
            files=None,
            headers=EXPECTED_HEADERS,
        )
        get_patched.assert_called_once_with(
            self.GET_METRICS_URL,
            json=None,
            params=self.BASIC_COMMAND_GET_METRICS_REQUEST_PARAMS,
            headers=EXPECTED_HEADERS,
        )
        assert result.exit_code == 0, result.exc_info

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_metrics_when_metrics_get_command_was_used_with_all_options(
            self, post_patched, get_patched):
        post_patched.return_value = MockResponse(self.GET_JOB_RESPONSE_JSON)
        get_patched.return_value = MockResponse(self.GET_METRICS_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.ALL_OPTIONS_COMMAND)

        # comparing objects instead of strings because Py2 and Py3 produce slightly different outputs
        assert json.loads(result.output.strip()) == json.loads(self.EXPECTED_STDOUT.strip()), result.exc_info
        post_patched.assert_called_once_with(
            self.GET_JOB_URL,
            json=self.GET_JOB_REQUEST_JSON,
            params=None,
            data=None,
            files=None,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
        )
        get_patched.assert_called_once_with(
            self.GET_METRICS_URL,
            json=None,
            params=self.ALL_COMMANDS_GET_METRICS_REQUEST_PARAMS,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
        )

        assert result.exit_code == 0, result.exc_info

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_metrics_when_metrics_get_was_executed_and_options_file_was_used(
            self, post_patched, get_patched, jobs_metrics_get_config_path):
        post_patched.return_value = MockResponse(self.GET_JOB_RESPONSE_JSON)
        get_patched.return_value = MockResponse(self.GET_METRICS_RESPONSE_JSON)

        command = self.FULL_OPTIONS_COMMAND_WITH_OPTIONS_FILE[:] + [jobs_metrics_get_config_path]
        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        # comparing objects instead of strings because Py2 and Py3 produce slightly different outputs
        assert json.loads(result.output.strip()) == json.loads(self.EXPECTED_STDOUT.strip()), result.exc_info
        post_patched.assert_called_once_with(
            self.GET_JOB_URL,
            json=self.GET_JOB_REQUEST_JSON,
            params=None,
            data=None,
            files=None,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
        )
        get_patched.assert_called_once_with(
            self.GET_METRICS_URL,
            json=None,
            params=self.ALL_COMMANDS_GET_METRICS_REQUEST_PARAMS,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
        )
        assert result.exit_code == 0, result.exc_info

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_valid_error_message_when_invalid_api_key_was_used(self, post_patched):
        post_patched.return_value = MockResponse({"status": 400, "message": "Invalid API token"}, status_code=400)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.ALL_OPTIONS_COMMAND)

        assert result.output == self.EXPECTED_STDOUT_WHEN_INVALID_API_KEY_WAS_USED, result.exc_info

        post_patched.assert_called_once_with(
            self.GET_JOB_URL,
            json=self.GET_JOB_REQUEST_JSON,
            params=None,
            data=None,
            files=None,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
        )

        assert result.exit_code == 0, result.exc_info

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_valid_error_message_when_job_was_not_found(self, post_patched):
        post_patched.return_value = MockResponse(self.GET_JOB_RESPONSE_JSON_WHEN_NO_JOBS_WERE_FOUND, status_code=400)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.ALL_OPTIONS_COMMAND)

        assert result.output == self.EXPECTED_STDOUT_WHEN_EXPERIMENT_WAS_NOT_FOUND, result.exc_info

        post_patched.assert_called_once_with(
            self.GET_JOB_URL,
            json=self.GET_JOB_REQUEST_JSON,
            params=None,
            data=None,
            files=None,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
        )
        assert result.exit_code == 0, result.exc_info

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_valid_message_when_was_no_metrics_were_returned(
            self, post_patched, get_patched):
        post_patched.return_value = MockResponse(self.GET_JOB_RESPONSE_JSON)
        get_patched.return_value = MockResponse(example_responses.JOBS_METRICS_GET_RESPONSE_WHEN_NO_DATA_WAS_FOUND)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.ALL_OPTIONS_COMMAND)

        assert json.loads(result.output.strip()) == json.loads(self.EXPECTED_STDOUT_WHEN_NO_METRICS_FOUND.strip()), \
            result.exc_info

        post_patched.assert_called_once_with(
            self.GET_JOB_URL,
            json=self.GET_JOB_REQUEST_JSON,
            params=None,
            data=None,
            files=None,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
        )
        get_patched.assert_called_once_with(
            self.GET_METRICS_URL,
            json=None,
            params=self.ALL_COMMANDS_GET_METRICS_REQUEST_PARAMS,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
        )

        assert result.exit_code == 0, result.exc_info

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_valid_error_message_when_error_code_was_returned_without_error_message(
            self, post_patched, get_patched):
        post_patched.return_value = MockResponse(self.GET_JOB_RESPONSE_JSON)
        get_patched.return_value = MockResponse(status_code=500)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.ALL_OPTIONS_COMMAND)

        assert result.output == self.EXPECTED_STDOUT_WHEN_ERROR_CODE_WAS_RETURNED_WITHOUT_ERROR_MESSAGE, result.exc_info
        post_patched.assert_called_once_with(
            self.GET_JOB_URL,
            json=self.GET_JOB_REQUEST_JSON,
            params=None,
            data=None,
            files=None,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
        )
        get_patched.assert_called_once_with(
            self.GET_METRICS_URL,
            json=None,
            params=self.ALL_COMMANDS_GET_METRICS_REQUEST_PARAMS,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
        )

        assert result.exit_code == 0, result.exc_info


class TestJobsMetricsStreamCommand(object):
    GET_JOB_URL = "https://api.paperspace.io/jobs/getPublicJob"
    GET_METRICS_URL = "https://aws-testing.paperspace.io/metrics/api/v1/stream"
    BASIC_OPTIONS_COMMAND = [
        "jobs", "metrics", "stream",
        "--id", "jstkd2lapucirs",
    ]
    ALL_OPTIONS_COMMAND = [
        "jobs", "metrics", "stream",
        "--id", "jstkd2lapucirs",
        "--metric", "gpuMemoryFree",
        "--metric", "gpuMemoryUsed",
        "--interval", "20s",
        "--apiKey", "some_key",
    ]
    ALL_OPTIONS_COMMAND_WITH_OPTIONS_FILE = [
        "jobs", "metrics", "stream",
        "--optionsFile",  # path added in test,
    ]

    GET_JOB_REQUEST_JSON = {"jobId": "jstkd2lapucirs"}
    BASIC_COMMAND_CHART_DESCRIPTOR = '{"chart_names": ["cpuPercentage", "memoryUsage"], "handles": ["jstkd2lapucirs"]' \
                                     ', "object_type": "mljob", "poll_interval": "30s"}'
    ALL_COMMANDS_CHART_DESCRIPTOR = '{"chart_names": ["gpuMemoryFree", "gpuMemoryUsed"], "handles": ["jstkd2lapucirs"' \
                                    '], "object_type": "mljob", "poll_interval": "20s"}'

    GET_JOB_RESPONSE_JSON = example_responses.GET_JOB_RESPONSE

    EXPECTED_TABLE_1 = """+------------------------------+---------------+-------------+
| Pod                          | cpuPercentage | memoryUsage |
+------------------------------+---------------+-------------+
| mljob-ecgrgm7ok8chv-0-worker |               | 5881856     |
+------------------------------+---------------+-------------+"""
    EXPECTED_TABLE_2 = """+------------------------------+---------------+-------------+
| Pod                          | cpuPercentage | memoryUsage |
+------------------------------+---------------+-------------+
| mljob-ecgrgm7ok8chv-0-worker | 0             | 5881856     |
+------------------------------+---------------+-------------+"""
    EXPECTED_TABLE_3 = """+------------------------------+---------------+-------------+
| Pod                          | cpuPercentage | memoryUsage |
+------------------------------+---------------+-------------+
| mljob-ecgrgm7ok8chv-0-worker | 0             | 5881857     |
+------------------------------+---------------+-------------+"""

    ALL_OPTIONS_EXPECTED_TABLE_1 = """+------------------------------+---------------+---------------+
| Pod                          | gpuMemoryFree | gpuMemoryUsed |
+------------------------------+---------------+---------------+
| mljob-ecgrgm7ok8chv-0-worker |               | 5881856       |
+------------------------------+---------------+---------------+"""
    ALL_OPTIONS_EXPECTED_TABLE_2 = """+------------------------------+---------------+---------------+
| Pod                          | gpuMemoryFree | gpuMemoryUsed |
+------------------------------+---------------+---------------+
| mljob-ecgrgm7ok8chv-0-worker | 0             | 5881856       |
+------------------------------+---------------+---------------+"""
    ALL_OPTIONS_EXPECTED_TABLE_3 = """+------------------------------+---------------+---------------+
| Pod                          | gpuMemoryFree | gpuMemoryUsed |
+------------------------------+---------------+---------------+
| mljob-ecgrgm7ok8chv-0-worker | 0             | 5881857       |
+------------------------------+---------------+---------------+"""

    EXPECTED_STDOUT_WHEN_INVALID_API_KEY_WAS_USED = "Failed to fetch data: Incorrect API Key provided\nForbidden\n"
    EXPECTED_STDOUT_WHEN_JOB_WAS_NOT_FOUND = "Failed to fetch data: No such job\n"

    @mock.patch("gradient.commands.common.TerminalPrinter")
    @mock.patch("gradient.api_sdk.repositories.common.websocket.create_connection")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_all_available_metrics_when_metrics_get_command_was_used_with_basic_options(
            self, post_patched, create_ws_connection_patched, terminal_printer_cls_patched,
            basic_options_metrics_stream_websocket_connection_iterator):
        post_patched.return_value = MockResponse(self.GET_JOB_RESPONSE_JSON)

        ws_connection_instance_mock = mock.MagicMock()
        ws_connection_instance_mock.__iter__ = basic_options_metrics_stream_websocket_connection_iterator
        create_ws_connection_patched.return_value = ws_connection_instance_mock

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.BASIC_OPTIONS_COMMAND)

        terminal_printer_cls_patched().init.assert_called_once()
        terminal_printer_cls_patched().rewrite_screen.assert_has_calls([
            mock.call(self.EXPECTED_TABLE_1),
            mock.call(self.EXPECTED_TABLE_2),
            mock.call(self.EXPECTED_TABLE_3),
        ])
        terminal_printer_cls_patched().cleanup.assert_called_once()

        post_patched.assert_called_once_with(
            self.GET_JOB_URL,
            json=self.GET_JOB_REQUEST_JSON,
            params=None,
            data=None,
            files=None,
            headers=EXPECTED_HEADERS,
        )
        ws_connection_instance_mock.send.assert_called_once_with(self.BASIC_COMMAND_CHART_DESCRIPTOR)
        assert result.exit_code == 0, result.exc_info

    @mock.patch("gradient.commands.common.TerminalPrinter")
    @mock.patch("gradient.api_sdk.repositories.common.websocket.create_connection")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_metrics_when_metrics_get_command_was_used_with_all_options(
            self, post_patched, create_ws_connection_patched, terminal_printer_cls_patched,
            all_options_metrics_stream_websocket_connection_iterator):
        post_patched.return_value = MockResponse(self.GET_JOB_RESPONSE_JSON)

        ws_connection_instance_mock = mock.MagicMock()
        ws_connection_instance_mock.__iter__ = all_options_metrics_stream_websocket_connection_iterator
        create_ws_connection_patched.return_value = ws_connection_instance_mock

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.ALL_OPTIONS_COMMAND)

        terminal_printer_cls_patched().init.assert_called_once()
        terminal_printer_cls_patched().rewrite_screen.assert_has_calls([
            mock.call(self.ALL_OPTIONS_EXPECTED_TABLE_1),
            mock.call(self.ALL_OPTIONS_EXPECTED_TABLE_2),
            mock.call(self.ALL_OPTIONS_EXPECTED_TABLE_3),
        ])
        terminal_printer_cls_patched().cleanup.assert_called_once()

        post_patched.assert_called_once_with(
            self.GET_JOB_URL,
            json=self.GET_JOB_REQUEST_JSON,
            params=None,
            data=None,
            files=None,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
        )

        ws_connection_instance_mock.send.assert_called_once_with(self.ALL_COMMANDS_CHART_DESCRIPTOR)
        assert result.exit_code == 0, result.exc_info

    @mock.patch("gradient.commands.common.TerminalPrinter")
    @mock.patch("gradient.api_sdk.repositories.common.websocket.create_connection")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_metrics_when_metrics_get_was_executed_and_options_file_was_used(
            self, post_patched, create_ws_connection_patched, terminal_printer_cls_patched,
            all_options_metrics_stream_websocket_connection_iterator,
            jobs_metrics_stream_config_path):
        post_patched.return_value = MockResponse(self.GET_JOB_RESPONSE_JSON)
        ws_connection_instance_mock = mock.MagicMock()
        ws_connection_instance_mock.__iter__ = all_options_metrics_stream_websocket_connection_iterator
        create_ws_connection_patched.return_value = ws_connection_instance_mock

        command = self.ALL_OPTIONS_COMMAND_WITH_OPTIONS_FILE[:] + [jobs_metrics_stream_config_path]
        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        terminal_printer_cls_patched().init.assert_called_once()
        terminal_printer_cls_patched().rewrite_screen.assert_has_calls([
            mock.call(self.ALL_OPTIONS_EXPECTED_TABLE_1),
            mock.call(self.ALL_OPTIONS_EXPECTED_TABLE_2),
            mock.call(self.ALL_OPTIONS_EXPECTED_TABLE_3),
        ])
        terminal_printer_cls_patched().cleanup.assert_called_once()

        post_patched.assert_called_once_with(
            self.GET_JOB_URL,
            json=self.GET_JOB_REQUEST_JSON,
            params=None,
            data=None,
            files=None,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
        )

        ws_connection_instance_mock.send.assert_called_once_with(self.ALL_COMMANDS_CHART_DESCRIPTOR)
        assert result.exit_code == 0, result.exc_info

    @mock.patch("gradient.commands.common.TerminalPrinter")
    @mock.patch("gradient.api_sdk.repositories.common.websocket.create_connection")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_valid_error_message_when_invalid_api_key_was_used(
            self, post_patched, create_ws_connection_patched, terminal_printer_cls_patched):
        post_patched.return_value = MockResponse({"status": 400, "message": "Invalid API token"}, 400)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.ALL_OPTIONS_COMMAND)

        assert "Failed to fetch data: Invalid API token\n" == result.output, result.exc_info

        post_patched.assert_called_once_with(
            self.GET_JOB_URL,
            json=self.GET_JOB_REQUEST_JSON,
            params=None,
            data=None,
            files=None,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
        )

        create_ws_connection_patched.assert_not_called()
        assert result.exit_code == 0, result.exc_info

    @mock.patch("gradient.commands.common.TerminalPrinter")
    @mock.patch("gradient.api_sdk.repositories.common.websocket.create_connection")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_valid_error_message_when_job_was_not_found(
            self, post_patched, create_ws_connection_patched, terminal_printer_cls_patched):
        post_patched.return_value = MockResponse(
            {"error": {"name": "ApplicationError", "status": 404, "message": "No such job"}},
            404,
        )

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.ALL_OPTIONS_COMMAND)

        assert result.output == self.EXPECTED_STDOUT_WHEN_JOB_WAS_NOT_FOUND, result.exc_info

        post_patched.assert_called_once_with(
            self.GET_JOB_URL,
            json=self.GET_JOB_REQUEST_JSON,
            params=None,
            data=None,
            files=None,
            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
        )

        create_ws_connection_patched.assert_not_called()
        assert result.exit_code == 0, result.exc_info



from .base_client import BaseClient, TagsSupportMixin
from .. import models, repositories


class ProjectsClient(TagsSupportMixin, BaseClient):
    entity = "project"

    def create(self, name, repository_name=None, repository_url=None, tags=None, ):
        """Create new project

        *EXAMPLE*::

            gradient projects create --name new-project

        *EXAMPLE RETURN*::

            Project created with ID: <your-project-id>


        in sdk::

            from gradient.api_sdk.clients import ProjectsClient

            api_key = 'your-api-key'
            projects_client = ProjectsClient(api_key)

            new_project = projects_client.create('your-project-name')

            print(new_project)

        :param str name: Name of new project [required]
        :param str repository_name: Name of the repository
        :param str repository_url: URL to the repository
        :param list[str] tags: List of tags

        :returns: project ID
        :rtype: str
        """

        project = models.Project(
            name=name,
            repository_name=repository_name,
            repository_url=repository_url,
        )

        repository = self.build_repository(repositories.CreateProject)
        handle = repository.create(project)

        if tags:
            self.add_tags(entity_id=handle, tags=tags)

        return handle

    def list(self, tags=None):
        """Get list of your projects

        :param list[str]|tuple[str] tags: tags to filter with OR

        :returns: list of projects
        :rtype: list[models.Project]
        """

        repository = self.build_repository(repositories.ListProjects)
        projects = repository.list(tags=tags)
        return projects

    def delete(self, project_id):
        repository = self.build_repository(repositories.DeleteProject)
        repository.delete(project_id)

    def get(self, project_id):
        repository = self.build_repository(repositories.GetProject)
        project = repository.get(id=project_id)
        return project



import json

import marshmallow

from .base import BaseSchema
from .. import models


class JSONField(marshmallow.fields.Field):
    """Field that serializes to json
    """

    def _serialize(self, value, attr, obj, **kwargs):
        if value is None:
            return None

        return json.dumps(value)

    def _deserialize(self, value, attr, data, **kwargs):
        return value


class JobDatasetSchema(BaseSchema):
    MODEL = models.JobDataset

    id = marshmallow.fields.Str(required=True)
    name = marshmallow.fields.Str(required=True)
    output = marshmallow.fields.Bool()


class JobSchema(BaseSchema):
    MODEL = models.Job

    id = marshmallow.fields.Str(dump_to="id", load_from="id")
    name = marshmallow.fields.Str(required=True)
    state = marshmallow.fields.Str()
    workspace_file_name = marshmallow.fields.Str(dump_to="workspaceFileName", load_from="workspaceFileName")
    working_directory = marshmallow.fields.Str(dump_to="workingDirectory", load_from="workingDirectory")
    artifacts_directory = marshmallow.fields.Str(dump_to="artifactsDirectory", load_from="artifactsDirectory")
    entrypoint = marshmallow.fields.Str()
    project_id = marshmallow.fields.Str(dump_to="projectId", load_from="projectId")
    project = marshmallow.fields.Str()
    container = marshmallow.fields.Str()
    container_url = marshmallow.fields.Str(dump_to="containerUrl", load_from="containerUrl")
    base_container = marshmallow.fields.Str(dump_to="baseContainer", load_from="baseContainer")
    base_container_url = marshmallow.fields.Str(dump_to="baseContainerUrl", load_from="baseContainerUrl")
    machine_type = marshmallow.fields.Str(dump_to="machineType", load_from="machineType")
    cluster = marshmallow.fields.Str()
    cluster_id = marshmallow.fields.Str(dump_to="clusterId", load_from="clusterId")
    usage_rate = marshmallow.fields.Str(dump_to="usageRate", load_from="usageRate")
    started_by_user_id = marshmallow.fields.Str(dump_to="startedByUserId", load_from="startedByUserId")
    parent_job_id = marshmallow.fields.Str(dump_to="parentJobId", load_from="parentJobId")
    job_error = marshmallow.fields.Str(dump_to="jobError", load_from="jobError")
    dt_created = marshmallow.fields.Str(dump_to="dtCreated", load_from="dtCreated")
    dt_modified = marshmallow.fields.Str(dump_to="dtModified", load_from="dtModified")
    dt_provisioning_started = marshmallow.fields.Str(dump_to="dtProvisioningStarted", load_from="dtProvisioningStarted")
    dt_provisioning_finished = marshmallow.fields.Str(dump_to="dtProvisioningFinished",
                                                      load_from="dtProvisioningFinished")
    dt_started = marshmallow.fields.Str(dump_to="dtStarted", load_from="dtStarted")
    dt_finished = marshmallow.fields.Str(dump_to="dtFinished", load_from="dtFinished")
    dt_teardown_started = marshmallow.fields.Str(dump_to="dtTeardownStarted", load_from="dtTeardownStarted")
    dt_teardown_finished = marshmallow.fields.Str(dump_to="dtTeardownFinished", load_from="dtTeardownFinished")
    dt_deleted = marshmallow.fields.Str(dump_to="dtDeleted", load_from="dtDeleted")
    exit_code = marshmallow.fields.Str(dump_to="exitCode", load_from="exitCode")
    queue_position = marshmallow.fields.Str(dump_to="queuePosition", load_from="queuePosition")
    seq_num = marshmallow.fields.Int(dump_to="seqNum", load_from="seqNum")
    storage_region = marshmallow.fields.Str(dump_to="storageRegion", load_from="storageRegion")
    cluster_machine = marshmallow.fields.Str(dump_to="clusterMachine", load_from="clusterMachine")
    fqdn = marshmallow.fields.Str()
    ports = marshmallow.fields.Str()
    is_public = marshmallow.fields.Bool(dump_to="isPublic", load_from="isPublic")
    container_user = marshmallow.fields.Str(dump_to="containerUser", load_from="containerUser")
    has_code = marshmallow.fields.Bool(dump_to="hasCode", load_from="hasCode")
    code_uploaded = marshmallow.fields.Bool(dump_to="codeUploaded", load_from="codeUploaded")
    code_commit = marshmallow.fields.Str(dump_to="codeCommit", load_from="codeCommit")
    run_till_cancelled = marshmallow.fields.Bool(dump_to="runTillCancelled", load_from="runTillCancelled")
    push_on_completion = marshmallow.fields.Bool(dump_to="pushOnCompletion", load_from="pushOnCompletion")
    new_image_name = marshmallow.fields.Str(dump_to="newImageName", load_from="newImageName")
    cpu_hostname = marshmallow.fields.Str(dump_to="cpuHostname", load_from="cpuHostname")
    cpu_count = marshmallow.fields.Int(dump_to="cpuCount", load_from="cpuCount")
    cpu_model = marshmallow.fields.Str(dump_to="cpuModel", load_from="cpuModel")
    cpu_flags = marshmallow.fields.Str(dump_to="cpuFlags", load_from="cpuFlags")
    cpu_mem = marshmallow.fields.Str(dump_to="cpuMem", load_from="cpuMem")
    gpu_name = marshmallow.fields.Str(dump_to="gpuName", load_from="gpuName")
    gpu_serial = marshmallow.fields.Str(dump_to="gpuSerial", load_from="gpuSerial")
    gpu_device = marshmallow.fields.Str(dump_to="gpuDevice", load_from="gpuDevice")
    gpu_driver = marshmallow.fields.Str(dump_to="gpuDriver", load_from="gpuDriver")
    gpu_count = marshmallow.fields.Int(dump_to="gpuCount", load_from="gpuCount")
    gpu_mem = marshmallow.fields.Str(dump_to="gpuMem", load_from="gpuMem")
    tpu_type = marshmallow.fields.Str(dump_to="tpuType", load_from="tpuType")
    tpu_name = marshmallow.fields.Str(dump_to="tpuName", load_from="tpuName")
    tpu_grpc_url = marshmallow.fields.Str(dump_to="tpuGrpcUrl", load_from="tpuGrpcUrl")
    tpu_tf_version = marshmallow.fields.Str(dump_to="tpuTFVersion", load_from="tpuTFVersion")
    tpu_dataset_dir = marshmallow.fields.Str(dump_to="tpuDatasetDir", load_from="tpuDatasetDir")
    tpu_model_dir = marshmallow.fields.Str(dump_to="tpuModelDir", load_from="tpuModelDir")
    target_node_attrs = marshmallow.fields.Dict(dump_to="targetNodeAttrs", load_from="targetNodeAttrs")
    job_env = marshmallow.fields.Dict(dump_to="jobEnv", load_from="jobEnv")
    env_vars = JSONField(dump_to="envVars", load_from="envVars")
    shared_mem_mbytes = marshmallow.fields.Int(dump_to="sharedMemMBytes", load_from="sharedMemMBytes")
    shutdown_timeout = marshmallow.fields.Int(dump_to="shutdownTimeout", load_from="shutdownTimeout")
    is_preemptible = marshmallow.fields.Bool(dump_to="isPreemptible", load_from="isPreemptible")
    metrics_url = marshmallow.fields.Str(dump_to="metricsURL", load_from="metricsURL")
    custom_metrics = marshmallow.fields.Str(dump_to="customMetrics", load_from="customMetrics")
    experiment_id = marshmallow.fields.Str(dump_to="experimentId", load_from="experimentId")

    command = marshmallow.fields.Str()
    use_dockerfile = marshmallow.fields.Bool(dump_to="useDockerfile", load_from="useDockerfile")
    rel_dockerfile_path = marshmallow.fields.Str(dump_to="relDockerfilePath", load_from="relDockerfilePath")
    registry_username = marshmallow.fields.Str(dump_to="registryUsername", load_from="registryUsername")
    registry_password = marshmallow.fields.Str(dump_to="registryPassword", load_from="registryPassword")
    build_only = marshmallow.fields.Bool(dump_to="buildOnly", load_from="buildOnly")

    registry_target = marshmallow.fields.Str(dump_to="registryTarget", load_from="registryTarget")
    registry_target_username = marshmallow.fields.Str(
        dump_to="registryTargetUsername", load_from="registryTargetUsername")
    registry_target_password = marshmallow.fields.Str(
        dump_to="registryTargetPassword", load_from="registryTargetPassword")

    tags = marshmallow.fields.Str(many=True, load_only=True)
    datasets = marshmallow.fields.Nested(JobDatasetSchema, dump_only=True, many=True)



from .base_client import BaseClient
from .. import repositories


class ClustersClient(BaseClient):
    def list(self, limit=11, offset=0, **kwargs):
        """
        Get a list of clusters for your team

        :param int limit: how many element to return on request
        :param int offset: from what position we should return clusters

        :return: clusters
        :rtype: list
        """
        repository = self.build_repository(repositories.ListClusters)
        clusters = repository.list(limit=limit, offset=offset)
        return clusters



import os
import shutil
import tempfile

import mock
from click.testing import CliRunner

from gradient.api_sdk.clients.http_client import default_headers
from gradient.cli import cli
from tests import example_responses, MockResponse
from tests.example_responses import LIST_MODEL_FILES_RESPONSE_JSON

EXPECTED_HEADERS = default_headers.copy()
EXPECTED_HEADERS["ps_client_name"] = "gradient-cli"

EXPECTED_HEADERS_WITH_CHANGED_API_KEY = EXPECTED_HEADERS.copy()
EXPECTED_HEADERS_WITH_CHANGED_API_KEY["X-API-Key"] = "some_key"


class TestModelsList(object):
    URL = "https://api.paperspace.io/mlModels/getModelList/"
    COMMAND = ["models", "list"]

    COMMAND_WITH_FILTERING_BY_EXPERIMENT_ID = [
        "models", "list",
        "--experimentId", "some_experiment_id",
        "--projectId", "some_project_id",
        "--tag", "some_tag",
        "--tag", "some_other_tag",
    ]
    EXPECTED_REQUEST_JSON_WITH_FILTERING = {"filter": {"where": {"and": [{"projectId": "some_project_id",
                                                                          "experimentId": "some_experiment_id"}]}},
                                            "tagFilter": ("some_tag", "some_other_tag")}

    COMMAND_WITH_API_KEY_PARAMETER_USED = ["models", "list", "--apiKey", "some_key"]
    COMMAND_WITH_OPTIONS_FILE = ["models", "list", "--optionsFile", ]  # path added in test

    EXPECTED_RESPONSE_JSON_WHEN_NO_MODELS_WERE_FOUND = {"modelList": [], "total": 1, "displayTotal": 0}

    EXPECTED_STDOUT = """+------+-----------------+------------+------------+---------------+
| Name | ID              | Model Type | Project ID | Experiment ID |
+------+-----------------+------------+------------+---------------+
| None | mosu30xm7q8vb0p | Tensorflow | prmr22ve0  | ehla1kvbwzaco |
+------+-----------------+------------+------------+---------------+
"""

    EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED = {"status": 401, "message": "No such API token"}

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_list_of_experiments(self, get_patched):
        get_patched.return_value = MockResponse(example_responses.LIST_MODELS_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params={"limit": -1})

        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_replate_api_key_in_headers_when_api_key_parameter_was_used(self, get_patched):
        get_patched.return_value = MockResponse(example_responses.LIST_MODELS_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_PARAMETER_USED)

        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                            json=None,
                                            params={"limit": -1})

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(self, get_patched, models_list_config_path):
        get_patched.return_value = MockResponse(example_responses.LIST_MODELS_RESPONSE_JSON)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [models_list_config_path]

        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                            json=self.EXPECTED_REQUEST_JSON_WITH_FILTERING,
                                            params={"limit": -1})

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_list_of_models_filtered_experiment_id(self, get_patched):
        get_patched.return_value = MockResponse(example_responses.LIST_MODELS_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_FILTERING_BY_EXPERIMENT_ID)

        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS,
                                            json=self.EXPECTED_REQUEST_JSON_WITH_FILTERING,
                                            params={"limit": -1})

        assert result.output == self.EXPECTED_STDOUT

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_proper_message_when_no_models_were_found(
            self, get_patched):
        get_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON_WHEN_NO_MODELS_WERE_FOUND)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params={"limit": -1})

        assert result.output == "No data found\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_proper_message_when_wrong_api_key_was_used(self, get_patched):
        get_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED, 401)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            params={"limit": -1})

        assert result.output == "Failed to fetch data: No such API token\n"


class TestDeleteModel(object):
    URL = "https://api.paperspace.io/mlModels/deleteModel/"
    COMMAND = ["models", "delete", "--id", "some_id"]
    EXPECTED_REQUEST_JSON = {"id": "some_id"}

    COMMAND_WITH_API_KEY_PARAMETER_USED = ["models", "delete", "--id", "some_id", "--apiKey", "some_key"]
    COMMAND_WITH_OPTIONS_FILE = ["models", "delete", "--id", "some_id", "--optionsFile", ]  # path added in test

    EXPECTED_STDOUT = "Model deleted\n"

    EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED = {"status": 400, "message": "Invalid API token"}

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_post_request_when_models_delete_command_was_executed(self, post_patched):
        post_patched.return_value = MockResponse(status_code=204)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             files=None,
                                             data=None,
                                             params=None)

        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_replace_api_key_in_headers_when_api_key_parameter_was_used(self, post_patched):
        post_patched.return_value = MockResponse(status_code=204)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_PARAMETER_USED)

        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             files=None,
                                             data=None,
                                             params=None)

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_options_from_yaml_file(self, post_patched, models_delete_config_path):
        post_patched.return_value = MockResponse(status_code=204)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [models_delete_config_path]

        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             files=None,
                                             data=None,
                                             params=None)

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_post_request_and_print_proper_message_when_model_with_given_id_was_not_found(
            self, post_patched):
        post_patched.return_value = MockResponse(example_responses.DELETE_MODEL_404_RESPONSE_JSON, status_code=404)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             files=None,
                                             data=None,
                                             params=None)

        assert result.output == "Failed to delete resource: Unable to find model\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_proper_message_when_wrong_api_key_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED, 401)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        post_patched.assert_called_once_with(self.URL,
                                             headers=EXPECTED_HEADERS,
                                             json=self.EXPECTED_REQUEST_JSON,
                                             files=None,
                                             data=None,
                                             params=None)

        assert result.output == "Failed to delete resource: Invalid API token\n"

class TestModelCreate(object):
    URL = "https://api.paperspace.io/mlModels/createModelV2"
    BASE_COMMAND = [
        "models", "create",
        "--name", "some_name",
        "--modelType", "custom",
        "--datasetRef", "dsr8k5qzn401lb5:latest",
    ]
    BASE_PARAMS = {
        "name": "some_name",
        "modelType": "Custom",
        "datasetRef": "dsr8k5qzn401lb5:latest"
    }
    COMMAND_WITH_ALL_OPTIONS = [
        "models", "create",
        "--name", "some_name",
        "--modelType", "tensorflow",
        "--modelSummary", """{"key": "value"}""",
        "--datasetRef", "dsr8k5qzn401lb5:latest",
        "--notes", "some notes",
        "--projectId", "some_project_id",
    ]
    ALL_OPTIONS_PARAMS = {
        "name": "some_name",
        "modelType": "Tensorflow",
        "summary": """{"key": "value"}""",
        "notes": "some notes",
        "projectId": "some_project_id",
        "datasetRef": "dsr8k5qzn401lb5:latest",
    }


    EXPECTED_STDOUT = "Model created with ID: some_model_id\n"

    CREATE_MODEL_V2_REPONSE = example_responses.MODEL_CREATE_RESPONSE_JSON_V2

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_post_request_when_models_create_command_was_used_with_basic_options(
            self, post_patched):
        post_patched.return_value = MockResponse(self.CREATE_MODEL_V2_REPONSE)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.BASE_COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_has_calls([
            mock.call(self.URL,
                        headers=EXPECTED_HEADERS,
                        json=None,
                        files=None,
                        data=None,
                        params=self.BASE_PARAMS),
        ])

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_post_request_when_models_update_command_was_used_with_all_options(
            self, post_patched):
        post_patched.return_value = MockResponse(self.CREATE_MODEL_V2_REPONSE)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_ALL_OPTIONS)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        post_patched.assert_has_calls([
            mock.call(self.URL,
                        headers=EXPECTED_HEADERS,
                        json=None,
                        files=None,
                        data=None,
                        params=self.ALL_OPTIONS_PARAMS),
        ])


class TestModelUpload(object):
    URL = "https://api.paperspace.io/mlModels/createModelV2"
    TAGS_URL = "https://api.paperspace.io/entityTags/updateTags"
    MODEL_FILE = "saved_model.pb"
    BASE_COMMAND = [
        "models", "upload",
        MODEL_FILE,
        "--name", "some_name",
        "--modelType", "custom",
    ]
    BASE_COMMAND_WITH_TAGS = [
        "models", "upload",
        MODEL_FILE,
        "--name", "some_name",
        "--modelType", "custom",
        "--tag", "test0",
        "--tag", "test1",
        "--tags", "test2,test3",
    ]
    BASE_PARAMS = {
        "name": "some_name",
        "modelType": "Custom",
    }
    TAGS_JSON = {
        "entity": "mlModel",
        "entityId": "some_model_id",
        "tags": ["test0", "test1", "test2", "test3"]
    }
    COMMAND_WITH_ALL_OPTIONS = [
        "models", "upload",
        MODEL_FILE,
        "--name", "some_name",
        "--modelType", "tensorflow",
        "--modelSummary", """{"key": "value"}""",
        "--notes", "some notes",
        "--projectId", "some_project_id",
        "--clusterId", "some_cluster_id",
    ]
    ALL_OPTIONS_PARAMS = {
        "name": "some_name",
        "modelType": "Tensorflow",
        "summary": """{"key": "value"}""",
        "notes": "some notes",
        "projectId": "some_project_id",
    }

    COMMAND_WITH_API_KEY_PARAMETER_USED = [
        "models", "upload",
        MODEL_FILE,
        "--name", "some_name",
        "--modelType", "tensorflow",
        "--modelSummary", """{"key": "value"}""",
        "--notes", "some notes",
        "--projectId", "some_project_id",
        "--clusterId", "some_cluster_id",
        "--apiKey", "some_key",
    ]
    COMMAND_WITH_OPTIONS_FILE = ["models", "upload", "--optionsFile", ]  # path added in test

    EXPECTED_STDOUT = "Model uploaded with ID: some_model_id\n"

    GET_PRESIGNED_URL = "https://api.paperspace.io/mlModels/getPresignedModelUrl"
    GET_PRESIGNED_URL_PARAMS = {"fileName": "saved_model.pb", "modelHandle": "some_model_id", "contentType": "", "clusterId": "some_cluster_id"}
    GET_PRESIGNED_URL_PARAMS_BASIC = {"fileName": "saved_model.pb", "modelHandle": "some_model_id", "contentType": ""}
    GET_PRESIGNED_URL_RESPONSE = example_responses.MODEL_UPLOAD_GET_PRESIGNED_URL_RESPONSE

    CREATE_MODEL_V2_REPONSE = example_responses.MODEL_CREATE_RESPONSE_JSON_V2

    EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED = {"status": 400, "message": "Invalid API token"}
    UPDATE_TAGS_RESPONSE_JSON_200 = example_responses.UPDATE_TAGS_RESPONSE

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_post_request_when_models_update_command_was_used_with_basic_options(
            self, post_patched, put_patched, get_patched):
        post_patched.return_value = MockResponse(self.CREATE_MODEL_V2_REPONSE)
        put_patched.return_value = MockResponse()
        get_patched.return_value = MockResponse(self.GET_PRESIGNED_URL_RESPONSE)

        runner = CliRunner()
        with runner.isolated_filesystem():
            with open(self.MODEL_FILE, "w") as h:
                h.write("I'm a model!")

            result = runner.invoke(cli.cli, self.BASE_COMMAND)

            assert result.output == self.EXPECTED_STDOUT, result.exc_info
            post_patched.assert_has_calls([
                mock.call(self.URL,
                          headers=EXPECTED_HEADERS,
                          json=None,
                          files=None,
                          data=None,
                          params=self.BASE_PARAMS),
            ])
            put_patched.assert_has_calls([
                mock.call(self.GET_PRESIGNED_URL_RESPONSE,
                          headers={"Content-Type": mock.ANY},
                          json=None,
                          params=None,
                          data=mock.ANY)
            ])
            get_patched.assert_called_once_with(self.GET_PRESIGNED_URL,
                                                headers=EXPECTED_HEADERS,
                                                params=self.GET_PRESIGNED_URL_PARAMS_BASIC,
                                                json=None,
                                                )
            assert put_patched.call_args.kwargs["data"].encoder.fields["file"][0] == self.MODEL_FILE

            assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_post_request_when_models_update_command_was_used_with_all_options(
            self, post_patched, put_patched, get_patched):
        post_patched.return_value = MockResponse(self.CREATE_MODEL_V2_REPONSE)
        put_patched.return_value = MockResponse()
        get_patched.return_value = MockResponse(self.GET_PRESIGNED_URL_RESPONSE)

        runner = CliRunner()
        with runner.isolated_filesystem():
            with open(self.MODEL_FILE, "w") as h:
                h.write("I'm a model!")

            result = runner.invoke(cli.cli, self.COMMAND_WITH_ALL_OPTIONS)

            assert result.output == self.EXPECTED_STDOUT, result.exc_info
            post_patched.assert_has_calls([
                mock.call(self.URL,
                          headers=EXPECTED_HEADERS,
                          json=None,
                          files=None,
                          data=None,
                          params=self.ALL_OPTIONS_PARAMS),
            ])
            put_patched.assert_has_calls([
                mock.call(self.GET_PRESIGNED_URL_RESPONSE,
                          headers={"Content-Type": mock.ANY},
                          json=None,
                          params=None,
                          data=mock.ANY)])
            get_patched.assert_called_once_with(self.GET_PRESIGNED_URL,
                                                headers=EXPECTED_HEADERS,
                                                params=self.GET_PRESIGNED_URL_PARAMS,
                                                json=None,
                                                )
            assert put_patched.call_args.kwargs["data"].encoder.fields["file"][0] == self.MODEL_FILE

            assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_replace_api_key_in_headers_when_api_key_parameter_was_used(
            self, post_patched, put_patched, get_patched):
        post_patched.return_value = MockResponse(self.CREATE_MODEL_V2_REPONSE)
        put_patched.return_value = MockResponse()
        get_patched.return_value = MockResponse(self.GET_PRESIGNED_URL_RESPONSE)

        runner = CliRunner()
        with runner.isolated_filesystem():
            with open(self.MODEL_FILE, "w") as h:
                h.write("I'm a model!")

            result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_PARAMETER_USED)

            assert result.output == self.EXPECTED_STDOUT, result.exc_info
            post_patched.assert_has_calls([
                mock.call(self.URL,
                          headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                          json=None,
                          files=None,
                          data=None,
                          params=self.ALL_OPTIONS_PARAMS
                          ),
            ])
            put_patched.assert_has_calls([
                mock.call(self.GET_PRESIGNED_URL_RESPONSE,
                          headers={"Content-Type": mock.ANY},
                          json=None,
                          params=None,
                          data=mock.ANY)
            ])
            get_patched.assert_called_once_with(self.GET_PRESIGNED_URL,
                                                headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                                params=self.GET_PRESIGNED_URL_PARAMS,
                                                json=None,
                                                )
            assert put_patched.call_args.kwargs["data"].encoder.fields["file"][0] == self.MODEL_FILE

            assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_read_options_from_yaml_file(
            self, post_patched, put_patched, get_patched, models_upload_config_path):
        post_patched.return_value = MockResponse(self.CREATE_MODEL_V2_REPONSE)
        put_patched.return_value = MockResponse()
        get_patched.return_value = MockResponse(self.GET_PRESIGNED_URL_RESPONSE)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [models_upload_config_path]

        runner = CliRunner()
        with runner.isolated_filesystem():
            with open(self.MODEL_FILE, "w") as h:
                h.write("I'm a model!")

            result = runner.invoke(cli.cli, command)

            assert result.output == self.EXPECTED_STDOUT, result.exc_info
            post_patched.assert_has_calls([
                mock.call(
                    self.URL,
                    headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                    json=None,
                    files=None,
                    data=None,
                    params=self.ALL_OPTIONS_PARAMS
                ),
            ])
            put_patched.assert_called_once_with(self.GET_PRESIGNED_URL_RESPONSE,
                                                headers={"Content-Type": ""},
                                                json=None,
                                                params=None,
                                                data=mock.ANY)
            get_patched.assert_called_once_with(self.GET_PRESIGNED_URL,
                                                headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                                params=self.GET_PRESIGNED_URL_PARAMS,
                                                json=None,
                                                )
            assert put_patched.call_args.kwargs["data"].encoder.fields["file"][0] == self.MODEL_FILE

            assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_print_proper_message_when_wrong_api_key_was_used(self, post_patched):
        post_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED, 401)

        runner = CliRunner()
        with runner.isolated_filesystem():
            with open(self.MODEL_FILE, "w") as h:
                h.write("I'm a model!")

            result = runner.invoke(cli.cli, self.BASE_COMMAND)

            assert result.output == "Failed to create resource: Invalid API token\n", result.exc_info
            post_patched.assert_called_with(self.URL,
                                            headers=EXPECTED_HEADERS,
                                            json=None,
                                            files=None,
                                            data=None,
                                            params=self.BASE_PARAMS)

            assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.put")
    @mock.patch("gradient.api_sdk.clients.http_client.requests.post")
    def test_should_send_proper_data_and_tag_machine(
            self, post_patched, put_patched, get_patched):
        post_patched.return_value = MockResponse(self.CREATE_MODEL_V2_REPONSE)
        put_patched.return_value = MockResponse()
        get_patched.side_effect = [MockResponse(self.GET_PRESIGNED_URL_RESPONSE),
                                   MockResponse({})]

        runner = CliRunner()
        with runner.isolated_filesystem():
            with open(self.MODEL_FILE, "w") as h:
                h.write("I'm a model!")

            result = runner.invoke(cli.cli, self.BASE_COMMAND_WITH_TAGS)
            assert result.output == self.EXPECTED_STDOUT, result.exc_info
            post_patched.assert_has_calls([
                mock.call(self.URL,
                          headers=EXPECTED_HEADERS,
                          json=None,
                          files=None,
                          data=None,
                          params=self.BASE_PARAMS
                          ),
            ])

            get_patched.assert_has_calls(
                [
                    mock.call(
                        self.GET_PRESIGNED_URL,
                        headers=EXPECTED_HEADERS,
                        params=self.GET_PRESIGNED_URL_PARAMS_BASIC,
                        json=None,
                    ),
                ]
            )
            put_patched.assert_has_calls(
                [
                    mock.call(
                        self.GET_PRESIGNED_URL_RESPONSE,
                        headers={"Content-Type": mock.ANY},
                        json=None,
                        params=None,
                        data=mock.ANY,
                    ),
                    mock.call(
                        self.TAGS_URL,
                        headers=EXPECTED_HEADERS,
                        json=self.TAGS_JSON,
                        params=None,
                        data=None,
                    ),
                ]
            )

            assert EXPECTED_HEADERS["X-API-Key"] != "some_key"


class TestModelDetails(object):
    URL = "https://api.paperspace.io/mlModels/getModelList/"
    COMMAND = ["models", "details", "--id", "some_id"]

    EXPECTED_REQUEST_JSON = {"filter": {"where": {"and": [{"id": "some_id"}]}}}

    COMMAND_WITH_API_KEY_PARAMETER_USED = ["models", "details", "--id", "some_id", "--apiKey", "some_key"]
    COMMAND_WITH_OPTIONS_FILE = ["models", "details", "--optionsFile", ]  # path added in test

    EXPECTED_RESPONSE_JSON_WHEN_NO_MODELS_WERE_FOUND = {"modelList": [], "total": 1, "displayTotal": 0}

    EXPECTED_STDOUT = """+------------------+----------------------------------------------------------------------------+
| ID               | some_id                                                                    |
+------------------+----------------------------------------------------------------------------+
| Name             | some_name                                                                  |
| Project ID       | some_project_id                                                            |
| Experiment ID    | some_experiment_id                                                         |
| Model Type       | Tensorflow                                                                 |
| URL              | s3://ps-projects-development/asdf/some_project_id/some_experiment_id/model |
| Deployment State | Stopped                                                                    |
| Tags             |                                                                            |
+------------------+----------------------------------------------------------------------------+
"""

    EXPECTED_STDOUT_WITH_TAGS = """+------------------+----------------------------------------------------------------------------+
| ID               | some_id                                                                    |
+------------------+----------------------------------------------------------------------------+
| Name             | some_name                                                                  |
| Project ID       | some_project_id                                                            |
| Experiment ID    | some_experiment_id                                                         |
| Model Type       | Tensorflow                                                                 |
| URL              | s3://ps-projects-development/asdf/some_project_id/some_experiment_id/model |
| Deployment State | Stopped                                                                    |
| Tags             | tag1, tag2                                                                 |
+------------------+----------------------------------------------------------------------------+
"""

    EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED = {"status": 400, "message": "Invalid API token"}

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_details_of_experiment(self, get_patched):
        get_patched.return_value = MockResponse(example_responses.MODEL_DETAILS_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT, result.exc_info
        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS,
                                            json=self.EXPECTED_REQUEST_JSON,
                                            params=None)

        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_details_of_experiment_that_has_some_tags(self, get_patched):
        get_patched.return_value = MockResponse(example_responses.MODEL_DETAILS_RESPONSE_JSON_WITH_TAGS)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        assert result.output == self.EXPECTED_STDOUT_WITH_TAGS, result.exc_info
        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS,
                                            json=self.EXPECTED_REQUEST_JSON,
                                            params=None)

        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_replace_api_key_in_headers_when_api_key_parameter_was_used(self, get_patched):
        get_patched.return_value = MockResponse(example_responses.MODEL_DETAILS_RESPONSE_JSON)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND_WITH_API_KEY_PARAMETER_USED)

        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                            json=self.EXPECTED_REQUEST_JSON,
                                            params=None)

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_read_options_from_yaml_file(self, get_patched, models_details_config_path):
        get_patched.return_value = MockResponse(example_responses.MODEL_DETAILS_RESPONSE_JSON)
        command = self.COMMAND_WITH_OPTIONS_FILE[:] + [models_details_config_path]

        runner = CliRunner()
        result = runner.invoke(cli.cli, command)

        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS_WITH_CHANGED_API_KEY,
                                            json=self.EXPECTED_REQUEST_JSON,
                                            params=None)

        assert result.output == self.EXPECTED_STDOUT
        assert EXPECTED_HEADERS["X-API-Key"] != "some_key"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_send_get_request_and_print_proper_message_when_no_models_were_found(
            self, get_patched):
        get_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_JSON_WHEN_NO_MODELS_WERE_FOUND)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS,
                                            json=self.EXPECTED_REQUEST_JSON,
                                            params=None)

        assert result.output == "Model not found\n"

    @mock.patch("gradient.api_sdk.clients.http_client.requests.get")
    def test_should_print_proper_message_when_wrong_api_key_was_used(self, get_patched):
        get_patched.return_value = MockResponse(self.EXPECTED_RESPONSE_WHEN_WRONG_API_KEY_WAS_USED, 401)

        runner = CliRunner()
        result = runner.invoke(cli.cli, self.COMMAND)

        get_patched.assert_called_once_with(self.URL,
                                            headers=EXPECTED_HEADERS,
                                            json=self.EXPECTED_REQUEST_JSON,
                                            params=None)

        assert result.output == "Failed to fetch data: Invalid API token\n"


class TestDownloadModelFiles(object):
    runner = CliRunner()
    LIST_FILES_URL = "https://api.paperspace.io/mlModels/listFiles/"
    DESTINATION_DIR_NAME = "dest"
    DESTINATION_DIR_PATH = os.path.join(tempfile.gettempdir(), "dest")

    COMMAND = ["models", "download", "--id", "some_model_id", "--destinationDir", DESTINATION_DIR_PATH]

    @classmethod
    def teardown_method(cls):
        shutil.rmtree(cls.DESTINATION_DIR_PATH)

    @mock.patch("gradient.api_sdk.s3_downloader.requests.get")
    def test_should_get_a_list_of_files_and_download_them_to_defined_directory_when_download_command_was_executed(
            self, get_patched,
    ):
        file_response_mock = mock.MagicMock()
        file_response_mock.content = "\"Hello Paperspace!\n\""
        file_response_mock_2 = mock.MagicMock()
        file_response_mock_2.content = "\"Hello Paperspace 2\n\""
        file_response_mock_3 = mock.MagicMock()
        file_response_mock_3.content = "\"Elo\n\""
        get_patched.side_effect = [
            MockResponse(LIST_MODEL_FILES_RESPONSE_JSON),
            file_response_mock,
            file_response_mock_2,
            file_response_mock_3,
        ]

        result = self.runner.invoke(cli.cli, self.COMMAND)

        get_patched.assert_has_calls([
            mock.call(self.LIST_FILES_URL,
                      headers=EXPECTED_HEADERS,
                      json={"links": True, "id": "some_model_id"},
                      params=None),
            mock.call("https://ps-projects.s3.amazonaws.com/some/path/model/hello.txt?AWSAccessKeyId="
                      "some_aws_access_key_id&Expires=713274132&Signature=7CT5k6buEmZe5k5E7g6BXMs2xV4%3D&"
                      "response-content-disposition=attachment%3Bfilename%3D%22hello.txt%22&x-amz-security-token="
                      "some_amz_security_token"),
            mock.call("https://ps-projects.s3.amazonaws.com/some/path/model/hello2.txt?AWSAccessKeyId="
                      "some_aws_access_key_id&Expires=713274132&Signature=L1lI47cNyiROzdYkf%2FF3Cm3165E%3D&"
                      "response-content-disposition=attachment%3Bfilename%3D%22hello2.txt%22&x-amz-security-token="
                      "some_amz_security_token"),
            mock.call("https://ps-projects.s3.amazonaws.com/some/path/model/keton/elo.txt?AWSAccessKeyId="
                      "some_aws_access_key_id&Expires=713274132&Signature=tHriojGx03S%2FKkVGQGVI5CQRFTo%3D&"
                      "response-content-disposition=attachment%3Bfilename%3D%22elo.txt%22&x-amz-security-token="
                      "some_amz_security_token"),
        ])
        assert os.path.exists(self.DESTINATION_DIR_PATH)
        assert os.path.isdir(self.DESTINATION_DIR_PATH)
        assert os.path.exists(os.path.join(self.DESTINATION_DIR_PATH, "keton"))
        assert os.path.isdir(os.path.join(self.DESTINATION_DIR_PATH, "keton"))

        hello_txt_path = os.path.join(self.DESTINATION_DIR_PATH, "hello.txt")
        assert os.path.exists(hello_txt_path)
        assert not os.path.isdir(hello_txt_path)
        with open(hello_txt_path) as h:
            assert h.read() == "\"Hello Paperspace!\n\""

        hello2_txt_path = os.path.join(self.DESTINATION_DIR_PATH, "hello2.txt")
        assert os.path.exists(hello2_txt_path)
        assert not os.path.isdir(hello2_txt_path)
        with open(hello2_txt_path) as h:
            assert h.read() == "\"Hello Paperspace 2\n\""

        elo_txt_path = os.path.join(self.DESTINATION_DIR_PATH, "keton", "elo.txt")
        assert os.path.exists(elo_txt_path)
        assert not os.path.isdir(elo_txt_path)
        with open(elo_txt_path) as h:
            assert h.read() == "\"Elo\n\""

        assert result.exit_code == 0



from .base_client import BaseClient, TagsSupportMixin
from .. import repositories, models
from ..repositories.machines import CheckMachineAvailability, DeleteMachine, ListMachines, WaitForState


class MachinesClient(TagsSupportMixin, BaseClient):
    entity = "machine"

    def create(
            self,
            name,
            machine_type,
            region,
            size,
            billing_type,
            template_id,
            assign_public_ip=None,
            dynamic_public_ip=None,
            network_id=None,
            team_id=None,
            user_id=None,
            email=None,
            password=None,
            first_name=None,
            last_name=None,
            notification_email=None,
            script_id=None,
            tags=None,
    ):
        """Create new machine

        :param str name: A memorable name for this machine [required]
        :param str machine_type: Machine type  [required]
        :param str region: Name of the region  [required]
        :param str size: Storage size for the machine in GB [required]
        :param str billing_type: Either 'monthly' or 'hourly' billing [required]
        :param str template_id: Template id of the template to use for creating this machine  [required]
        :param bool assign_public_ip: Assign a new public ip address. Cannot be used with dynamic_public_ip
        :param bool dynamic_public_ip: Temporarily assign a new public ip address on machine.
                                       Cannot be used with assign_public_ip
        :param str network_id: If creating on a specific network, specify its id
        :param str team_id: If creating the machine for a team, specify the team id
        :param str user_id: If assigning to an existing user other than yourself, specify the user id
                            (mutually exclusive with email, password, first_name, last_name)
        :param str email: If creating a new user for this machine, specify their email address
                          (mutually exclusive with user_id)
        :param str password: If creating a new user, specify their password (mutually exclusive with user_id)
        :param str first_name: If creating a new user, specify their first name (mutually exclusive with user_id)
        :param str last_name: If creating a new user, specify their last name (mutually exclusive with user_id)
        :param str notification_email: Send a notification to this email address when complete
        :param str script_id: The script id of a script to be run on startup
        :param list[str] tags: List of tags

        :returns: ID of created machine
        :rtype: str
        """

        instance = models.Machine(
            name=name,
            machine_type=machine_type,
            region=region,
            size=size,
            billing_type=billing_type,
            template_id=template_id,
            assign_public_ip=assign_public_ip,
            dynamic_public_ip=dynamic_public_ip,
            network_id=network_id,
            team_id=team_id,
            user_id=user_id,
            email=email,
            password=password,
            first_name=first_name,
            last_name=last_name,
            notification_email=notification_email,
            script_id=script_id,
        )

        repository = self.build_repository(repositories.CreateMachine)
        handle = repository.create(instance)
        if tags:
            self.add_tags(entity_id=handle, tags=tags)
        return handle

    def get(self, id):
        """Get machine instance

        :param str id: ID of a machine [required]

        :return: Machine instance
        :rtype: models.Machine
        """
        repository = self.build_repository(repositories.GetMachine)
        instance = repository.get(id=id)
        return instance

    def is_available(self, machine_type, region):
        """Check if specified machine is available in certain region

        :param str machine_type: Machine type  [required]
        :param str region: Name of the region  [required]

        :return: If specified machine is available in the region
        :rtype: bool
        """

        repository = self.build_repository(CheckMachineAvailability)
        handle = repository.get(machine_type=machine_type, region=region)
        return handle

    def restart(self, id):
        """Restart machine

        :param str id: ID of a machine [required]
        """

        repository = self.build_repository(repositories.RestartMachine)
        repository.restart(id)

    def start(self, id):
        """Start machine

        :param str id: ID of a machine [required]
        """

        repository = self.build_repository(repositories.StartMachine)
        repository.start(id)

    def stop(self, id):
        """Stop machine

        :param str id: ID of a machine [required]
        """

        repository = self.build_repository(repositories.StopMachine)
        repository.stop(id)

    def update(
            self,
            id,
            name=None,
            shutdown_timeout_in_hours=None,
            shutdown_timeout_forces=None,
            perform_auto_snapshot=None,
            auto_snapshot_frequency=None,
            auto_snapshot_save_count=None,
            dynamic_public_ip=None,
    ):
        """Update machine instance

        :param str id: Id of the machine to update  [required]
        :param str name: New name for the machine
        :param int shutdown_timeout_in_hours: Number of hours before machine is shutdown if no one is logged in
                                              via the Paperspace client
        :param bool shutdown_timeout_forces: Force shutdown at shutdown timeout, even if there is
                                             a Paperspace client connection
        :param bool perform_auto_snapshot: Perform auto snapshots
        :param str auto_snapshot_frequency: One of 'hour', 'day', 'week', or None
        :param int auto_snapshot_save_count: Number of snapshots to save
        :param str dynamic_public_ip: If true, assigns a new public ip address on machine start and releases it
                                      from the account on machine stop
        """
        instance = models.Machine(
            name=name,
            dynamic_public_ip=dynamic_public_ip,
            shutdown_timeout_in_hours=shutdown_timeout_in_hours,
            shutdown_timeout_forces=shutdown_timeout_forces,
            perform_auto_snapshot=perform_auto_snapshot,
            auto_snapshot_frequency=auto_snapshot_frequency,
            auto_snapshot_save_count=auto_snapshot_save_count,
        )

        repository = self.build_repository(repositories.UpdateMachine)
        repository.update(id, instance)

    def get_utilization(self, id, billing_month):
        """

        :param id: ID of the machine
        :param billing_month: Billing month in "YYYY-MM" format

        :return: Machine utilization info
        :rtype: models.MachineUtilization
        """
        repository = self.build_repository(repositories.GetMachineUtilization)
        usage = repository.get(id=id, billing_month=billing_month)
        return usage

    def delete(self, machine_id, release_public_ip=False):
        """Destroy machine with given ID

        :param str machine_id: ID of the machine
        :param bool release_public_ip: If the assigned public IP should be released
        """

        repository = self.build_repository(DeleteMachine)
        repository.delete(machine_id, release_public_ip=release_public_ip)

    def wait_for_state(self, machine_id, state, interval=5):
        """Wait for defined machine state

        :param str machine_id: ID of the machine
        :param str state: State of machine to wait for
        :param int interval: interval between polls
        """

        repository = self.build_repository(WaitForState)
        repository.wait_for_state(machine_id, state, interval)

    def list(
            self,
            id=None,
            name=None,
            os=None,
            ram=None,
            cpus=None,
            gpu=None,
            storage_total=None,
            storage_used=None,
            usage_rate=None,
            shutdown_timeout_in_hours=None,
            perform_auto_snapshot=None,
            auto_snapshot_frequency=None,
            auto_snapshot_save_count=None,
            agent_type=None,
            created_timestamp=None,
            state=None,
            updates_pending=None,
            network_id=None,
            private_ip_address=None,
            public_ip_address=None,
            region=None,
            user_id=None,
            team_id=None,
            last_run_timestamp=None,
    ):
        """

        :param str id: Optional machine id to match on
        :param str name: Filter by machine name
        :param str os: Filter by os used
        :param int ram: Filter by machine RAM (in bytes)
        :param int cpus: Filter by CPU count
        :param str gpu: Filter by GPU type
        :param str storage_total: Filter by total storage
        :param str storage_used: Filter by storage used
        :param str usage_rate: Filter by usage rate
        :param int shutdown_timeout_in_hours: Filter by shutdown timeout
        :param bool perform_auto_snapshot: Filter by performAutoSnapshot flag
        :param str auto_snapshot_frequency: Filter by autoSnapshotFrequency flag
        :param int auto_snapshot_save_count: Filter by auto shapshots count
        :param str agent_type: Filter by agent type
        :param datetime created_timestamp: Filter by date created
        :param str state: Filter by state
        :param str updates_pending: Filter by updates pending
        :param str network_id: Filter by network ID
        :param str private_ip_address: Filter by private IP address
        :param str public_ip_address: Filter by public IP address
        :param str region: Filter by region. One of {CA, NY2, AMS1}
        :param str user_id: Filter by user ID
        :param str team_id: Filter by team ID
        :param str last_run_timestamp: Filter by last run date

        :return: List of machines
        :rtype: list[models.Machine]
        """

        repository = self.build_repository(ListMachines)
        machines = repository.list(
            id=id,
            name=name,
            os=os,
            ram=ram,
            cpus=cpus,
            gpu=gpu,
            storage_total=storage_total,
            storage_used=storage_used,
            usage_rate=usage_rate,
            shutdown_timeout_in_hours=shutdown_timeout_in_hours,
            perform_auto_snapshot=perform_auto_snapshot,
            auto_snapshot_frequency=auto_snapshot_frequency,
            auto_snapshot_save_count=auto_snapshot_save_count,
            agent_type=agent_type,
            created_timestamp=created_timestamp,
            state=state,
            updates_pending=updates_pending,
            network_id=network_id,
            private_ip_address=private_ip_address,
            public_ip_address=public_ip_address,
            region=region,
            user_id=user_id,
            team_id=team_id,
            last_run_timestamp=last_run_timestamp,
        )
        return machines



import datetime

import attr

from ..config import config
from ..utils import concatenate_urls


@attr.s
class Notebook(object):
    id = attr.ib(type=str, default=None)
    machine_type = attr.ib(type=str, default=None)
    vm_type_id = attr.ib(type=int, default=None)
    vm_type_label = attr.ib(type=str, default=None)
    container_id = attr.ib(type=int, default=None)
    container_name = attr.ib(type=str, default=None)
    name = attr.ib(type=str, default=None)
    cluster_id = attr.ib(type=str, default=None)
    registry_username = attr.ib(type=str, default=None)
    registry_password = attr.ib(type=str, default=None)
    command = attr.ib(type=str, default=None)
    default_entrypoint = attr.ib(type=str, default=None)
    container_user = attr.ib(type=str, default=None)
    shutdown_timeout = attr.ib(type=int, default=None)
    is_preemptible = attr.ib(type=bool, default=None)
    project_id = attr.ib(type=str, default=None)
    project_handle = attr.ib(type=str, default=None)
    state = attr.ib(type=str, default=None)
    vm_type = attr.ib(type=str, default=None)
    fqdn = attr.ib(type=str, default=None)
    namespace = attr.ib(type=str, default=None)
    tags = attr.ib(type=list, factory=list)
    metrics_url = attr.ib(type=str, default=None)
    is_public = attr.ib(type=str, default=None)
    token = attr.ib(type=str, default=None)
    job_error = attr.ib(type=str, default=None)
    job_handle = attr.ib(type=str, default=None)
    container = attr.ib(type=str, default=None)
    container_url = attr.ib(type=str, default=None)
    base_container = attr.ib(type=str, default=None)
    base_container_url = attr.ib(type=str, default=None)
    environment = attr.ib(type=dict, default=None)
    workspace = attr.ib(type=str, default=None)
    workspace_username = attr.ib(type=str, default=None)
    workspace_password = attr.ib(type=str, default=None)
    workspace_ref = attr.ib(type=str, default=None)
    should_run_on_create = attr.ib(type=str, default=None)

    dt_created = attr.ib(type=datetime.datetime, default=None)
    dt_modified = attr.ib(type=datetime.datetime, default=None)
    dt_provisioning_started = attr.ib(type=datetime.datetime, default=None)
    dt_provisioning_finished = attr.ib(type=datetime.datetime, default=None)
    dt_started = attr.ib(type=datetime.datetime, default=None)
    dt_finished = attr.ib(type=datetime.datetime, default=None)
    dt_teardown_started = attr.ib(type=datetime.datetime, default=None)
    dt_teardown_finished = attr.ib(type=datetime.datetime, default=None)
    dt_workspace_upload_finished = attr.ib(type=datetime.datetime, default=None)
    dt_deleted = attr.ib(type=datetime.datetime, default=None)

    @property
    def url(self):
        url = concatenate_urls(config.WEB_URL, "/{}/notebook/{}".format(self.namespace, self.project_handle))
        return url


@attr.s
class NotebookStart(object):
    notebook_id = attr.ib(type=str, default=None)
    machine_type = attr.ib(type=str, default=None)
    vm_type_id = attr.ib(type=int, default=None)
    vm_type_label = attr.ib(type=str, default=None)
    cluster_id = attr.ib(type=str, default=None)
    shutdown_timeout = attr.ib(type=int, default=None)
    is_preemptible = attr.ib(type=bool, default=None)
    notebook_name = attr.ib(type=str, default=None)




import attr


@attr.s
class DatasetVersionTagSummary(object):
    """
    Dataset version tag summary class
    """
    name = attr.ib(type=str, default=None)


@attr.s
class DatasetVersion(object):
    """
    Dataset version class
    """
    version = attr.ib(type=str, default=None)
    message = attr.ib(type=str, default=None)
    is_committed = attr.ib(type=bool, default=None)
    tags = attr.ib(type=list, factory=list)

    # only used for create
    dataset_id = attr.ib(type=str, default=None)


@attr.s
class DatasetVersionPreSignedS3Call(object):
    """
    Dataset version pre-signed S3 call class
    """
    method = attr.ib(type=str, default=None)
    params = attr.ib(type=dict, factory=dict)


@attr.s
class DatasetVersionPreSignedURL(object):
    """
    Dataset version pre-signed URL class
    """
    url = attr.ib(type=str, default=None)
    expires_in = attr.ib(type=int, default=None)



import marshmallow

from .base import BaseSchema
from .. import models


class LogRowSchema(BaseSchema):
    MODEL = models.LogRow

    line = marshmallow.fields.Int()
    message = marshmallow.fields.Str()
    timestamp = marshmallow.fields.Str()

    @marshmallow.pre_load
    def rstrip_log_line(self, data):
        if "message" in data:
            data["message"] = str(data["message"]).rstrip()

        return data



import marshmallow

from .. import models
from .base import BaseSchema


class ArtifactSchema(BaseSchema):
    MODEL = models.Artifact

    file = marshmallow.fields.Str()
    url = marshmallow.fields.Str(required=True)
    size = marshmallow.fields.Int(required=True)
